/*  ppmmolw.c

    Project Earth image in Mollweide projection to rectangular
    Mercator bitmap.

    Based on the projection code from the GCTPc package, developed
    by the EROS Data Center.  GCTPc is available from the following
    FTP sites:

	ftp://edcftp.cr.usgs.gov/pub/software/gctpc/gctpc.tar.Z
	ftp://charon.er.usgs.gov/pub/S.gctpc.tar.Z

*/

#include <math.h>
#include "ppm.h"

#define PI	3.141592653589793238
#define TWO_PI	(PI*2.0)
#define EPSLN	1.0e-10
#define D2R	1.745329251994328e-2

#define MAX_VAL 4
#define MAXLONG 2147483647.
#define DBLLONG 4.61168601e18

/* Function to return the sign of an argument
  ------------------------------------------*/
static int sign(x)
  double x; 
{
    if (x < 0.0)
	return -1;
    else
	return 1;
}

/* Function to adjust a longitude angle to range from -180 to 180 radians
   added if statments 
  -----------------------------------------------------------------------*/
static double adjust_lon(x) 
  double x;		  /* Angle in radians			  */
{
    long temp;
    long count = 0;

    for(;;)
      {
      if (fabs(x)<=PI)
	 break;
      else
      if (((long) fabs(x / PI)) < 2)
	 x = x-(sign(x) *TWO_PI);
      else
      if (((long) fabs(x / TWO_PI)) < MAXLONG)
	 {
	 x = x-(((long)(x / TWO_PI))*TWO_PI);
	 }
      else
      if (((long) fabs(x / (MAXLONG * TWO_PI))) < MAXLONG)
	 {
	 x = x-(((long)(x / (MAXLONG * TWO_PI))) * (TWO_PI * MAXLONG));
	 }
      else
      if (((long) fabs(x / (DBLLONG * TWO_PI))) < MAXLONG)
	 {
	 x = x-(((long)(x / (DBLLONG * TWO_PI))) * (TWO_PI * DBLLONG));
	 }
      else
	 x = x-(sign(x) *TWO_PI);
      count++;
      if (count > MAX_VAL)
	 break;
      }

    return x;
}

/* Variables common to all subroutines in this code file
  -----------------------------------------------------*/
static double lon_center;	/* Center longitude (projection center) */
static double R;		/* Radius of the earth (sphere) */
static double false_easting;	/* x offset in meters			*/
static double false_northing;	/* y offset in meters			*/

/* Initialize the Mollweide projection
------------------------------------*/
static void molwforint(r, center_long,false_east,false_north)
  double r;			  /* (I) Radius of the earth (sphere)	  */
  double center_long;		  /* (I) Center longitude		  */
  double false_east;		  /* x offset in meters 		  */
  double false_north;		  /* y offset in meters 		  */
{
/* Place parameters in static storage for common use
  -------------------------------------------------*/
    false_easting = false_east;
    false_northing = false_north;
    R = r;
    lon_center = center_long;
    return;
}

/* Mollweide forward equations--mapping lat,long to x,y
  ----------------------------------------------------*/
static int molwfor(lon, lat, x, y)
  double lon;			  /* (I) Longitude */
  double lat;			  /* (I) Latitude */
  double *x;			  /* (O) X projection coordinate */
  double *y;			  /* (O) Y projection coordinate */
{
    double delta_lon;	    /* Delta longitude (Given longitude - center */
    double theta;
    double delta_theta;
    double con;
    long i;

    /* Forward equations
      -----------------*/
    delta_lon = adjust_lon(lon - lon_center);
    theta = lat;
    con = PI * sin(lat);

    /* Iterate using the Newton-Raphson method to find theta
      -----------------------------------------------------*/
    for (i=0;;i++)
       {
       delta_theta = -(theta + sin(theta) - con)/ (1.0 + cos(theta));
       theta += delta_theta;
       if (fabs(delta_theta) < EPSLN) break;
       if (i >= 50) 
	 {
         pm_error("Mollweide-forward iteration failed to converge: lon = %g lat = %g", lon, lat);
	 return(241);
	 }
       }
    theta /= 2.0;

    /* If the latitude is 90 deg, force the x coordinate to be "0 + false easting"
       this is done here because of precision problems with "cos(theta)"
       --------------------------------------------------------------------------*/
    if (PI/2 - fabs(lat) < EPSLN)
       delta_lon =0;
    *x = 0.900316316158 * R * delta_lon * cos(theta) + false_easting;
    *y = 1.4142135623731 * R * sin(theta) + false_northing;
    return 0;
}

int
main( argc, argv )
    int argc;
    char* argv[];
    {
    FILE* ifp, *afp = NULL;
    pixel ptab[256];
    pixel **ipixels, **opixels, **apixels;
    int cols, rows, acols, arows;
    pixval maxval, amaxval, r, g, b, lr, lg, lb;
    int argn;
    int i, j, ix, iy;
    double lat, lon, x, y;
    char* usage = 
        "[-dim <num>] [-red <num>] [-green <num>] [-blue <num>] [pbmfile]";


    ppm_init( &argc, argv );

    argn = 1;

    while ( argn < argc && argv[argn][0] == '-' && argv[argn][1] != '\0' )
	{
	    pm_usage( usage );
	++argn;
	}

    if ( argn != argc )
	{
	ifp = pm_openr( argv[argn] );
	++argn;
	}
    else
	ifp = stdin;

    if ( argn != argc )
	{
	afp = pm_openr( argv[argn] );
	++argn;
	}

    if ( argn != argc )
	pm_usage( usage );

    molwforint(1.0 / sqrt(2.0), 0.0, 0.0, 0.0);

    ipixels = ppm_readppm( ifp, &cols, &rows, &maxval );
    pm_close( ifp );
    if (afp != NULL) {
	apixels = ppm_readppm( afp, &acols, &arows, &amaxval );
	pm_close(afp);
	if (cols != acols || (cols / 2) != arows ||
	    maxval != amaxval) {
            pm_message("Auxiliary bitmap doesn't match input size");
	    afp = NULL;
	}
    }
    opixels = ppm_allocarray(cols, cols / 2);
    maxval = 255;

	for (i = 0; i < cols / 2; i++) {
	    for (j = 0; j < cols; j++) {
		lat = (90.0 * (i - (cols / 4))) / (cols / 4);
		lon = (180.0 * (j - (cols / 2))) / cols;

		molwfor(D2R * lon, D2R * lat, &x, &y);
		ix = (int) (x * (cols / 2)) + (cols / 2);
		iy = (int) (y * (cols / 4)) + (cols / 4);
		r = PPM_GETR(ipixels[iy][ix]);
		g = PPM_GETG(ipixels[iy][ix]);
		b = PPM_GETB(ipixels[iy][ix]);
		if (afp != NULL) {
		    int whitecomp, backcomp;
		    pixval ar, ag, ab;

		    ar = PPM_GETR(apixels[i][j]);
		    ag = PPM_GETG(apixels[i][j]);
		    ab = PPM_GETB(apixels[i][j]);
		    if (r != g || g != b) {
			r = lr;
			g = lg;
			b = lb;
		    }
		    lr = r;
		    lg = g;
		    lb = b;

		    whitecomp = r;
		    backcomp = (maxval - r);
		    r = ((whitecomp * maxval) + (backcomp * ar)) / maxval;
		    g = ((whitecomp * maxval) + (backcomp * ag)) / maxval;
		    b = ((whitecomp * maxval) + (backcomp * ab)) / maxval;
		} else {
		    if (r == 0 && g == 0 && b == 0) {
			r = g = (3 * maxval) / 4;
			b = maxval;
		    }
		}
		PPM_ASSIGN(opixels[i][j], r, g, b);

	    }
	}



    ppm_writeppm(stdout, opixels, cols, cols / 2, maxval, 0);
    pm_close(stdout);
    exit(0);
}
